/*
DCBouncer: bouncer / proxy for Direct Connect networks.
Copyright (C) 2008-2017 poy <poy@123gen.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "ADCHandler.h"

#include <set>

#include "dcpp/Encoder.h"
#include "dcpp/TigerHash.h"

#include "Client.h"
#include "Conversion.h"
#include "Hub.h"
#include "Math.h"

ADCHandler::ADCHandler(EventHandler& handler, const Profile& profile) :
Proxy(handler, profile)
{
}

void ADCHandler::Hub_info::clear()
{
	sup.clear();
	ssid.clear();
	sid = 0;
	infs.clear();
	cmds.clear();
}

void ADCHandler::checkHubSup()
{
	if (hub_info.sup.empty() || hub_info.ssid.empty())
	{
		return;
	}

	if (client_info.inf.get())
	{
		hub->Send(client_info.inf->toString(hub_info.sid));
	}
	else if (client_ok() && client->Ready())
	{
		client->Send(hub_info.sup);
		client->Send(hub_info.ssid);
	}
}

std::string ADCHandler::saveLogInf(const uint32_t sid)
{
	if (sid)
	{
		Hub_info::infs_type::const_iterator i = hub_info.infs.find(sid);
		if (i != hub_info.infs.end())
		{
			std::string cid;
			if (i->second->getParam("ID", 0, cid) && !cid.empty())
			{
				log_infs[cid] = i->second;
				return cid;
			}
		}
	}
	return std::string();
}

char ADCHandler::separator() const
{
	return '\n';
}

void ADCHandler::clientCommand(const std::string& line_)
{
	dcpp::AdcCommand command(line_);

	std::string line = line_ + '\n';

	bool forward = true;

	switch (command.getCommand())
	{
	case dcpp::AdcCommand::CMD_SUP:
		{
			size_t i = line.find(" ADZLIF");
			if (i != std::string::npos)
			{
				line.erase(i, line.find_first_of(" \n", i + 1) - i);
			}

			client_info.sup = line;
			save("sup", client_info.sup);

			if (client)
			{
				client_info.inf.reset();
			}
			else
			{
				// Connecting using cached data.
				std::string inf = load("inf");
				if (!inf.empty())
				{
					client_info.inf.reset(new dcpp::AdcCommand(inf));
				}

				std::string log_data = load("logs");
				i = 0;
				size_t j;
				uint8_t type = 0;
				Log log;
				while ((j = log_data.find('\n', i)) != std::string::npos)
				{
					std::string data = log_data.substr(i, j - i);
					i = j + 1;
					switch (type)
					{
					case 0:
						{
							log.message = data;
							++type;
							break;
						}
					case 1:
						{
							log.timestamp = atoi(data.c_str());
							++type;
							break;
						}
					case 2:
						{
							log.from = data;
							++type;
							break;
						}
					case 3:
						{
							log.to = data;
							++type;
							break;
						}
					case 4:
						{
							if (!log.message.empty())
							{
								log.pm = data;
								logs.push_back(log);
							}
							type = 0;
							break;
						}
					}
				}

				log_data = load("log_infs");
				i = 0;
				while ((j = log_data.find('\n', i)) != std::string::npos)
				{
					shared_command_type log_command(new dcpp::AdcCommand(log_data.substr(i, j - i)));
					i = j + 1;

					std::string cid;
					if (log_command->getParam("ID", 0, cid) && !cid.empty())
					{
						log_infs[cid] = log_command;
					}
				}
			}

			if (hub_ok() && hub->Ready())
			{
				checkHubSup();
			}

			forward = false;
			break;
		}

	case dcpp::AdcCommand::CMD_PAS:
		{
			forward = false;
			break;
		}

	case dcpp::AdcCommand::CMD_INF:
		{
			if (command.hasParam("ID"))
			{
				client_info.inf.reset(new dcpp::AdcCommand(line_));
				save("inf", line);

				// Determine whether this is the first connection to the hub.
				Hub_info::infs_type::const_iterator client_inf = hub_info.infs.find(hub_info.sid);
				bool reconnecting = client_inf != hub_info.infs.end();

				if (reconnecting)
				{
					std::string cid, nick, param;
					client_inf->second->getParam("ID", 0, cid);
					client_inf->second->getParam("NI", 0, nick);
					if ((command.getParam("ID", 0, param) && param != cid) ||
						(command.getParam("NI", 0, param) && param != nick))
					{
						Hub_info::infs_type::const_iterator hub_inf = hub_info.infs.find(dcpp::AdcCommand::HUB_SID);
						if (hub_inf != hub_info.infs.end())
						{
							client->Send(hub_inf->second->toString(hub_inf->first));
							client->Send(client_inf->second->toString(client_inf->first));
							sendMessage("CID / Nick change detected, reconnecting the hub...");
						}
						if (hub_ok())
						{
							hub->SetCloseAndDelete();
						}
						forward = false;
						break;
					}
				}

				for (Hub_info::infs_type::const_iterator i = hub_info.infs.begin(); i != hub_info.infs.end(); ++i)
				{
					client->Send(i->second->toString(i->first));
				}

				sendCommands();

				if (reconnecting)
				{
					// Make the command look like a regular INF.
					command.removeParam("ID");
					command.removeParam("PD");
					command.removeParam("NI");
					if (!profile.offline_description.empty() && !command.hasParam("DE"))
					{
						command.addParam("DE");
					}
					line = command.toString(hub_info.sid);

					for (Hub_info::cmds_type::const_iterator i = hub_info.cmds.begin(); i != hub_info.cmds.end(); ++i)
					{
						client->Send(*i);
					}

					if (!logs.empty())
					{
						// Find information on users mentioned in the messages.
						typedef std::map<std::string, uint32_t> sids_type;
						sids_type sids;
						typedef std::set<uint32_t> offline_sids_type;
						offline_sids_type offline_sids;
						for (log_infs_type::const_iterator i = log_infs.begin(); i != log_infs.end(); ++i)
						{
							uint32_t sid = 0;

							// Look into the list of online users.
							for (Hub_info::infs_type::const_iterator i_inf = hub_info.infs.begin(); i_inf != hub_info.infs.end(); ++i_inf)
							{
								std::string cid;
								if (i_inf->second->getParam("ID", 0, cid) && cid == i->first)
								{
									sid = i_inf->first;
									break;
								}
							}

							if (!sid)
							{
								// Generate a dummy SID.
								union
								{
									uint32_t sid;
									char chars[4];
								} sid_u;
								const size_t max = sizeof(dcpp::Encoder::base32Alphabet) - 1;
								do
								{
									sid_u.chars[0] = dcpp::Encoder::base32Alphabet[Math::rand(0, max)];
									sid_u.chars[1] = dcpp::Encoder::base32Alphabet[Math::rand(0, max)];
									sid_u.chars[2] = dcpp::Encoder::base32Alphabet[Math::rand(0, max)];
									sid_u.chars[3] = dcpp::Encoder::base32Alphabet[Math::rand(0, max)];
								}
								while (hub_info.infs.find(sid_u.sid) != hub_info.infs.end() ||
									offline_sids.find(sid_u.sid) != offline_sids.end());
								sid = sid_u.sid;

								// Send a temporary INF to pretend the user is online.
								client->Send(i->second->toString(sid));
								offline_sids.insert(sid);
							}

							sids[i->first] = sid;
						}
						log_infs.clear();

						while (!logs.empty())
						{
							logs_type::iterator i = logs.begin();

							dcpp::AdcCommand log_command(i->message);

							if (i->timestamp > 0)
							{
								log_command.addParam("TS", toString(i->timestamp));
							}

							sids_type::const_iterator i_sid = sids.find(i->from);
							uint32_t from;
							if (i_sid == sids.end())
							{
								log_command.setType(dcpp::AdcCommand::TYPE_INFO);
								from = 0;
							}
							else
							{
								from = i_sid->second;
							}

							if (!i->to.empty())
							{
								i_sid = sids.find(i->to);
								log_command.setTo((i_sid == sids.end()) ? 0 : i_sid->second);
							}
							if (!i->pm.empty())
							{
								i_sid = sids.find(i->pm);
								if (i_sid != sids.end())
								{
									log_command.removeParam("PM", 1);
									log_command.addParam("PM", dcpp::AdcCommand::fromSID(i_sid->second));
								}
							}

							client->Send(log_command.toString(from));
							logs.erase(i);
						}

						// Remove users that have been added although they were offline.
						for (offline_sids_type::const_iterator i = offline_sids.begin(); i != offline_sids.end(); ++i)
						{
							client->Send(dcpp::AdcCommand(dcpp::AdcCommand::CMD_QUI, dcpp::AdcCommand::TYPE_INFO)
								.addParam(dcpp::AdcCommand::fromSID(*i))
								.toString());
						}
					}

					save("logs", std::string());
					save("log_infs", std::string());
				}
			}
			break;
		}

	case dcpp::AdcCommand::CMD_CMD:
		{
			if (checkCommand(command.getParam(0)))
			{
				forward = false;
			}
			break;
		}
	}

	if (forward && hub_ok() && hub->Ready())
	{
		hub->Send(line);
	}
}

void ADCHandler::hubConnected()
{
	hub->Send(client_info.sup);
}

void ADCHandler::hubCommand(const std::string& line_)
{
	dcpp::AdcCommand command(line_);

	std::string line = line_ + '\n';

	bool forward = true;

	switch (command.getCommand())
	{
	case dcpp::AdcCommand::CMD_SUP:
		{
			if (command.getType() == dcpp::AdcCommand::TYPE_INFO)
			{
				size_t i = line.find(" ADZLIF");
				if (i != std::string::npos)
				{
					line.erase(i, line.find_first_of(" \n", i + 1) - i);
				}

				hub_info.sup = line;
				checkHubSup();
				forward = false;
			}
			break;
		}

	case dcpp::AdcCommand::CMD_SID:
		{
			if (command.getType() == dcpp::AdcCommand::TYPE_INFO)
			{
				hub_info.ssid = line;
				hub_info.sid = command.toSID(command.getParam(0));
				checkHubSup();
				forward = false;
			}
			break;
		}

	case dcpp::AdcCommand::CMD_INF:
		{
			uint32_t sid = command.getFrom();
			Hub_info::infs_type::iterator i = hub_info.infs.find(sid);
			if (i == hub_info.infs.end())
			{
				hub_info.infs.insert(std::make_pair(sid, shared_command_type(new dcpp::AdcCommand(line_))));
				if (sid == dcpp::AdcCommand::HUB_SID)
				{
					forward = false;
				}
			}
			else
			{
				i->second->parse(line_);
			}
			break;
		}

	case dcpp::AdcCommand::CMD_GPA:
		{
			if (command.getType() == dcpp::AdcCommand::TYPE_INFO)
			{
				std::string salt = command.getParam(0);
				if (!salt.empty())
				{
					size_t saltBytes = salt.size() * 5 / 8;
					std::vector<uint8_t> buf(saltBytes);
					dcpp::Encoder::fromBase32(salt.c_str(), &buf[0], saltBytes);
					dcpp::TigerHash th;
					th.update(profile.pass.data(), profile.pass.size());
					th.update(&buf[0], saltBytes);
					hub->Send(dcpp::AdcCommand(dcpp::AdcCommand::CMD_PAS, dcpp::AdcCommand::TYPE_HUB)
						.addParam(dcpp::Encoder::toBase32(th.finalize(), dcpp::TigerHash::BYTES))
						.toString(hub_info.sid));
				}
				forward = false;
			}
			break;
		}

	case dcpp::AdcCommand::CMD_CMD:
		{
			hub_info.cmds.push_back(line);
			break;
		}

	case dcpp::AdcCommand::CMD_QUI:
		{
			hub_info.infs.erase(command.toSID(command.getParam(0)));
			break;
		}

	case dcpp::AdcCommand::CMD_GET:
		if (command.getParam(0) == "blom")
		{
			forward = false;
		}
	}

	if (forward)
	{
		if (client_ok() && client->Ready())
		{
			client->Send(line);
		}

		else if (command.getCommand() == dcpp::AdcCommand::CMD_MSG)
		{
			Log log =
			{
				line_,
				time(0),
				saveLogInf(command.getFrom()),
				saveLogInf(command.getTo())
			};
			std::string pm;
			if (command.getParam("PM", 1, pm) && !pm.empty())
			{
				log.pm = saveLogInf(dcpp::AdcCommand::toSID(pm));
			}
			logs.push_back(log);

			save("logs",
				log.message + '\n' +
				toString(log.timestamp) + '\n' +
				log.from + '\n' +
				log.to + '\n' +
				log.pm + '\n', true);

			std::string data;
			for (log_infs_type::const_iterator i = log_infs.begin(); i != log_infs.end(); ++i)
			{
				data += i->second->toString(hub_info.sid);
			}
			save("log_infs", data);
		}
	}
}

void ADCHandler::clearHubInfo()
{
	hub_info.clear();
}

void ADCHandler::sendCommand(const std::string& text, const std::string& id)
{
	dcpp::AdcCommand command(dcpp::AdcCommand::CMD_CMD, dcpp::AdcCommand::TYPE_INFO);
	if (text.empty())
	{
		command.addParam(id).addParam("SP", "1");
	}
	else
	{
		command.addParam("DCBouncer/" + text)
			.addParam("TT", dcpp::AdcCommand(dcpp::AdcCommand::CMD_CMD, dcpp::AdcCommand::TYPE_HUB)
			.addParam(id).toString());
	}
	client->Send(command.addParam("CT", "1").toString());
}

void ADCHandler::sendMessage(const std::string& message)
{
	client->Send(dcpp::AdcCommand(dcpp::AdcCommand::CMD_MSG, dcpp::AdcCommand::TYPE_INFO)
		.addParam("<DCBouncer> " + message).toString());
}

void ADCHandler::offlineDescription(const std::string& offline_description)
{
	if (client_info.inf.get())
	{
		client_info.inf->removeParam("DE");
		client_info.inf->removeParam("SS");
		client_info.inf->addParam("DE", offline_description).addParam("SS", "0");
	}
	if (hub_ok() && hub->Ready())
	{
		hub->Send(dcpp::AdcCommand(dcpp::AdcCommand::CMD_INF, dcpp::AdcCommand::TYPE_BROADCAST)
			.addParam("DE", offline_description)
			.addParam("SS", "0")
			.toString(hub_info.sid));
	}
}
