/*
DCBouncer: bouncer / proxy for Direct Connect networks.
Copyright (C) 2008-2017 poy <poy@123gen.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Crypto.h"

#include <syslog.h>

#include <openssl/pem.h>

#include "FileSystem.h"

std::string Crypto::cert_file;

void Crypto::setCertFile(const std::string& cert_file_)
{
	cert_file = cert_file_;
}

const std::string& Crypto::getCertFile()
{
	if (cert_file.empty())
	{
		cert_file = FileSystem::absolute("../var/DCBouncer/cert.pem");
		FileSystem::ensureDir(cert_file);
		FILE* f = fopen(cert_file.c_str(), "w");
		if (!f)
		{
			syslog(LOG_WARNING, "can't get write access to the cert file (path: %s)", cert_file.c_str());
			cert_file.clear();
			return cert_file;
		}

		RSA* rsa = RSA_new();
		BIGNUM* bn = BN_new();
		EVP_PKEY* pkey = EVP_PKEY_new();
		X509* x509 = X509_new();
		ASN1_INTEGER* sn = ASN1_INTEGER_new();

		if (!(rsa && bn && pkey && x509 && sn &&
			BN_set_word(bn, RSA_F4) &&
			RSA_generate_key_ex(rsa, 2048, bn, 0) &&
			EVP_PKEY_set1_RSA(pkey, rsa) &&
			ASN1_INTEGER_set(sn, 1) &&
			X509_set_serialNumber(x509, sn) &&
			X509_gmtime_adj(X509_get_notBefore(x509), 0) &&
			X509_gmtime_adj(X509_get_notAfter(x509), 1095 * 24 * 3600) &&
			X509_set_pubkey(x509, pkey) &&
			X509_sign(x509, pkey, EVP_sha1()) &&
			PEM_write_RSAPrivateKey(f, rsa, 0, 0, 0, 0, 0) &&
			PEM_write_X509(f, x509)))
		{
			syslog(LOG_WARNING, "can't generate the contents of the cert file (path: %s)", cert_file.c_str());
			cert_file.clear();
		}

		if (rsa) { RSA_free(rsa); }
		if (bn) { BN_free(bn); }
		if (pkey) { EVP_PKEY_free(pkey); }
		if (x509) { X509_free(x509); }
		if (sn) { ASN1_INTEGER_free(sn); }

		fclose(f);
	}

	return cert_file;
}
