/*
DCBouncer: bouncer / proxy for Direct Connect networks.
Copyright (C) 2008-2017 poy <poy@123gen.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "EventHandler.h"

#include <fstream>

#include "Sockets/ListenSocket.h"

#include "ADCHandler.h"
#include "Client.h"
#include "Conversion.h"
#include "FileSystem.h"
#include "Hub.h"
#include "NMDCHandler.h"

EventHandler::EventHandler() :
sockets::EventHandler()
{
}

EventHandler::~EventHandler()
{
	for (initializers_type::iterator i = initializers.begin(); i != initializers.end(); ++i)
	{
		delete i->second;
	}
}

void EventHandler::add(sockets::Socket* socket, const Profile& profile)
{
	Initializer* initializer = new Initializer(*this, profile);
	initializers[profile.port] = initializer;

	if (secure(profile))
	{
		socket->EnableSSL();
	}
	Add(socket);

	// Simulate an ADC connection if a cached SUP is available.
	std::string sup = load(profile.port, "sup");
	if (!sup.empty())
	{
		initializer->clientConnected(0);
		initializer->clientLine(sup);
		initializer->clientDisconnected();
	}

	// Simulate an NMDC connection if a cached MyINFO is available.
	else if (!load(profile.port, "my_info").empty())
	{
		initializer->clientConnected(0);
		initializer->clientDisconnected();
	}
}

bool EventHandler::OkToAccept(sockets::Socket* p)
{
	for (socket_m::const_iterator i = m_sockets.begin(); i != m_sockets.end(); ++i)
	{
		Client* client = dynamic_cast<Client*>(i->second);
		if (client && client->GetParent() == p)
		{
			return false;
		}
	}
	return true;
}

void EventHandler::clientConnected(Client* client)
{
	port_t port = client->GetParent()->GetPort();

	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->clientConnected(client);
	}
}

void EventHandler::clientLine(port_t port, const std::string& line)
{
	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->clientLine(line);
	}
}

void EventHandler::clientDisconnected(port_t port)
{
	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->clientDisconnected();
	}
}

void EventHandler::hubConnected(port_t port)
{
	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->hubConnected();
	}
}

void EventHandler::hubLine(port_t port, const std::string& line)
{
	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->hubLine(line);
	}
}

void EventHandler::hubDisconnected(port_t port)
{
	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->hubDisconnected();
	}
}

void EventHandler::connectHub(port_t port)
{
	Initializer* initializer = getInitializer(port);
	if (initializer)
	{
		initializer->connectHub();
	}
}

std::string EventHandler::getDataFile(port_t port, const std::string& name) const
{
	return FileSystem::absolute("../var/DCBouncer/data/" + toString(port) + '/' + name);
}

std::string EventHandler::load(port_t port, const std::string& name) const
{
	std::ifstream f(getDataFile(port, name).c_str(), std::ios::in | std::ios::binary);
	f.seekg(0, std::ios::end);
	int size = f.tellg();
	if (size <= 0) { return std::string(); }
	f.seekg(0, std::ios::beg);
	char buf[size];
	f.read(buf, size);
	return std::string(buf, size);
}

void EventHandler::save(port_t port, const std::string& name, const std::string& data, bool append) const
{
	std::string file = getDataFile(port, name);
	FileSystem::ensureDir(file);
	std::ofstream(file.c_str(), std::ios::out | std::ios::binary | (append ? std::ios::app : std::ios::trunc))
		.write(data.c_str(), data.size());
}

EventHandler::Initializer::Initializer(EventHandler& handler, const Profile& profile) :
proxy(0),
handler(handler),
profile(profile),
client(0),
hub(0)
{
}

EventHandler::Initializer::~Initializer()
{
	if (proxy)
	{
		delete proxy;
	}
}

void EventHandler::Initializer::clientConnected(Client* client_)
{
	client = client_;

	if (proxy)
	{
		proxy->onClientConnected(client);
	}
	else if (!hub)
	{
		connectHub();
	}
}

void EventHandler::Initializer::clientLine(const std::string& line)
{
	if (!proxy)
	{
		if (line.compare(0, 4, "HSUP"))
		{
			// Invalid ADC handshake; most likely a TLS client.
			handler.setSecure(profile);
			client->GetParent()->EnableSSL();
			client->SetCloseAndDelete();
			if (hub_ok())
			{
				hub->SetCloseAndDelete();
			}
			return;
		}

		proxy = new ADCHandler(handler, profile);
		proxy->onClientConnected(client);
		if (hub_ok())
		{
			proxy->onHubConnected(hub);
		}
	}
	proxy->onClientLine(line);
}

void EventHandler::Initializer::clientDisconnected()
{
	client = 0;

	if (proxy)
	{
		proxy->onClientDisconnected();
	}
}

void EventHandler::Initializer::hubConnected()
{
	if (proxy)
	{
		proxy->onHubConnected(hub);
	}
}

void EventHandler::Initializer::hubLine(const std::string& line)
{
	if (!proxy)
	{
		proxy = new NMDCHandler(handler, profile);
		if (client_ok())
		{
			proxy->onClientConnected(client);
		}
		proxy->onHubConnected(hub);
	}
	proxy->onHubLine(line);
}

void EventHandler::Initializer::hubDisconnected()
{
	hub = 0;

	if (proxy)
	{
		proxy->onHubDisconnected();
	}

	if (client_ok())
	{
		client->SetCloseAndDelete();
	}
}

void EventHandler::Initializer::connectHub()
{
	if (proxy)
	{
		proxy->clearHubInfo();
	}
	if (hub_ok())
	{
		hub->SetCloseAndDelete();
	}

	hub = new Hub(handler, profile);
	if (handler.secure(profile))
	{
		hub->EnableSSL();
	}
	handler.Add(hub);
}

bool EventHandler::Initializer::client_ok() const
{
	return client && handler.Valid(client);
}

bool EventHandler::Initializer::hub_ok() const
{
	return hub && handler.Valid(hub);
}

EventHandler::Initializer* EventHandler::getInitializer(port_t port) const
{
	initializers_type::const_iterator i = initializers.find(port);
	if (i != initializers.end())
	{
		return i->second;
	}
	return 0;
}

bool EventHandler::secure(const Profile& profile) const
{
	return FileSystem::exists(getDataFile(profile.port, "secure"));
}

void EventHandler::setSecure(const Profile& profile) const
{
	save(profile.port, "secure", std::string(), false);
}
