/*
DCBouncer: bouncer / proxy for Direct Connect networks.
Copyright (C) 2008-2017 poy <poy@123gen.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "NMDCHandler.h"

#include "Client.h"
#include "Hub.h"

NMDCHandler::NMDCHandler(EventHandler& handler, const Profile& profile) :
Proxy(handler, profile)
{
	client_info.supports = load("supports");
	client_info.key = load("key");
	client_info.validate = load("validate");
	client_info.pass = load("pass");
	client_info.my_info = load("my_info");

	std::string log_data = load("logs");
	size_t i = 0, j;
	while ((j = log_data.find('|', i)) != std::string::npos)
	{
		logs.push_back(log_data.substr(i, j - i + 1));
		i = j + 1;
	}
}

void NMDCHandler::Client_info::clear()
{
	supports.clear();
	key.clear();
	validate.clear();
	// Don't clear the pass out since we don't ask for it again on new client connections.
	my_info.clear();
}

void NMDCHandler::Hub_info::clear()
{
	lock.clear();
	supports.clear();
	pass.clear();
	hello.clear();
	name.clear();
	cmds.clear();
	ips.clear();
}

void NMDCHandler::checkClientValidate()
{
	if (client_info.key.empty() || client_info.validate.empty())
	{
		return;
	}

	if (hub_info.hello.empty())
	{
		if (hub_ok() && hub->Ready())
		{
			hub->Send(client_info.key);
			hub->Send(client_info.validate);
		}
	}
	else
	{
		client->Send(hub_info.hello);
	}
}

char NMDCHandler::separator() const
{
	return '|';
}

void NMDCHandler::clientConnected()
{
	if (hub_ok() && hub->Ready())
	{
		client_info.clear();
		client->Send(hub_info.lock);
	}
}

void NMDCHandler::clientCommand(const std::string& line_)
{
	std::string line = line_ + '|';

	bool forward = true;

	std::string command;
	if (line.size() > 2 && line[0] == '$')
	{
		command = line.substr(1, line.find_first_of(" |") - 1);
	}

	if (command == "Supports")
	{
		size_t i = line.find(" ZPipe0");
		if (i != std::string::npos)
		{
			line.erase(i, line.find_first_of(" |", i + 1) - i);
		}

		client_info.supports = line;
		save("supports", client_info.supports);

		if (!hub_info.supports.empty())
		{
			client->Send(hub_info.supports);
			forward = false;
		}
	}

	else if (command == "Key")
	{
		client_info.key = line;
		save("key", client_info.key);
		checkClientValidate();
		forward = false;
	}

	else if (command == "ValidateNick")
	{
		client_info.validate = line;
		save("validate", client_info.validate);
		checkClientValidate();
		forward = false;
	}

	else if (command == "MyPass")
	{
		client_info.pass = line;
		save("pass", client_info.pass);

		if (!hub_info.hello.empty())
		{
			forward = false;
		}
	}

	else if (command == "Version")
	{
		forward = false;
	}

	else if (command == "GetNickList")
	{
		if (!hub_info.name.empty())
		{
			client->Send(hub_info.name);
		}

		sendCommands();

		for (Hub_info::cmds_type::const_iterator i = hub_info.cmds.begin(); i != hub_info.cmds.end(); ++i)
		{
			client->Send(*i);
		}

		while (!logs.empty())
		{
			logs_type::iterator i = logs.begin();
			client->Send(*i);
			logs.erase(i);
		}
		save("logs", std::string());
	}

	else if (command == "MyINFO")
	{
		client_info.my_info = line;
		save("my_info", client_info.my_info);
	}

	else if (checkCommand(command))
	{
		forward = false;
	}

	if (forward && hub_ok() && hub->Ready())
	{
		hub->Send(line);
	}
}

void NMDCHandler::hubCommand(const std::string& line_)
{
	std::string line = line_ + '|';

	bool forward = true;
	bool send_ips = false;

	std::string command;
	if (line.size() > 2 && line[0] == '$')
	{
		command = line.substr(1, line.find_first_of(" |") - 1);
	}

	if (command == "Lock")
	{
		hub_info.lock = line;

		if (!client_info.key.empty() && !client_info.validate.empty())
		{
			if (!client_info.supports.empty())
			{
				hub->Send(client_info.supports);
			}
			hub->Send(client_info.key);
			hub->Send(client_info.validate);
			forward = false;
		}
	}

	else if (command == "Supports")
	{
		hub_info.supports = line;

		if (!client_info.my_info.empty())
		{
			forward = false;
		}
	}

	else if (command == "GetPass")
	{
		hub_info.pass = line;

		if (!client_info.pass.empty())
		{
			hub->Send(client_info.pass);
			forward = false;
		}
	}

	else if (command == "Hello")
	{
		hub_info.hello = line;

		if (!client_info.my_info.empty())
		{
			hub->Send(client_info.my_info);
			forward = false;
		}
	}

	else if (command == "HubName")
	{
		hub_info.name = line;
	}

	else if (command == "UserCommand")
	{
		hub_info.cmds.push_back(line);
	}

	else if (command == "MyINFO")
	{
		if (line == client_info.my_info)
		{
			send_ips = true;
		}
	}

	else if (command == "UserIP")
	{
		hub_info.ips.push_back(line);
	}

	if (forward)
	{
		if (client_ok() && client->Ready())
		{
			client->Send(line);

			if (send_ips)
			{
				for (Hub_info::ips_type::const_iterator i = hub_info.ips.begin(); i != hub_info.ips.end(); ++i)
				{
					client->Send(*i);
				}
			}
		}

		else if (command.empty() || command == "To:")
		{
			logs.push_back(line);
			save("logs", line, true);
		}
	}
}

void NMDCHandler::clearHubInfo()
{
	hub_info.clear();
}

void NMDCHandler::sendCommand(const std::string& text, const std::string& id)
{
	bool separator = text.empty();
	std::string command = "$UserCommand ";
	command += separator ? '0' : '1';
	command += " 1 ";
	if (!separator)
	{
		command += "DCBouncer\\" + text + "$&#36;" + id + "&#124;";
	}
	client->Send(command + '|');
}

void NMDCHandler::sendMessage(const std::string& message)
{
	client->Send("<DCBouncer> " + message + '|');
}

void NMDCHandler::offlineDescription(const std::string& offline_description)
{
	if (client_info.my_info.size() > 20)
	{
		size_t space = client_info.my_info.find(' ', 13);
		if (space != std::string::npos)
		{
			size_t dollar = client_info.my_info.find('$', space);
			if (dollar != std::string::npos)
			{
				size_t tag = client_info.my_info.rfind('<', dollar);
				size_t end = (tag == std::string::npos) ? dollar : tag;
				space++;
				client_info.my_info.erase(space, end - space);
				client_info.my_info.insert(space, offline_description);
				dollar = client_info.my_info.rfind('$', client_info.my_info.size() - 3);
				if (dollar != std::string::npos)
				{
					dollar++;
					client_info.my_info.erase(dollar, client_info.my_info.size() - 2 - dollar);
					client_info.my_info.insert(dollar, "0");
				}
				if (hub_ok() && hub->Ready())
				{
					hub->Send(client_info.my_info);
				}
			}
		}
	}
}
