/*
DCBouncer: bouncer / proxy for Direct Connect networks.
Copyright (C) 2008-2017 poy <poy@123gen.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Proxy.h"

#include "Client.h"
#include "Conversion.h"
#include "Hub.h"
#include "main.h"

Proxy::Proxy(EventHandler& handler, const Profile& profile) :
sockets::IEventOwner(handler),
profile(profile),
client(0),
hub(0),
handler(handler),
hub_conn_event(0),
offline_event(0),
offline_percent(profile.offline_description.find('%')),
offline_hours(atoi(load("offline_hours").c_str()))
{
	commands.resize(Commands_last);
	commands[Reconnect] = "Reconnect the hub";
	commands[Restart] = "Restart DCBouncer";
	commands[Quit] = "Quit DCBouncer";
}

bool Proxy::client_ok() const
{
	return client && handler.Valid(client);
}

bool Proxy::hub_ok() const
{
	return hub && handler.Valid(hub);
}

std::string Proxy::load(const std::string& name) const
{
	return handler.load(profile.port, name);
}

void Proxy::save(const std::string& name, const std::string& data, bool append) const
{
	handler.save(profile.port, name, data, append);
}

void Proxy::sendCommands()
{
	for (unsigned i = Commands_first; i < Commands_last; ++i)
	{
		sendCommand(commands[i], "DCBouncer-" + toString(i));
	}
}

bool Proxy::checkCommand(const std::string& command)
{
	if (command.size() <= 10 || command.substr(0, 10) != "DCBouncer-")
	{
		return false;
	}
	switch (atoi(command.substr(10).c_str()))
	{
	case Reconnect:
		{
			sendMessage("Reconnecting the hub...");
			if (hub_ok())
			{
				hub->SetCloseAndDelete();
			}
			return true;
		}
	case Restart:
		{
			sendMessage("Restarting DCBouncer...");
			restart();
			return true;
		}
	case Quit:
		{
			sendMessage("Killing DCBouncer...");
			stop();
			return true;
		}
	}
	return false;
}

void Proxy::onClientConnected(Client* client_)
{
	client_buf.clear();
	client = client_;

	if (offline_percent != std::string::npos)
	{
		clearOfflineEvent();
	}

	clientConnected();
}

void Proxy::onClientLine(const std::string& line)
{
	parseCommands(client_buf, line, &Proxy::clientCommand);
}

void Proxy::onClientDisconnected()
{
	if (!profile.offline_description.empty())
	{
		if (offline_percent != std::string::npos)
		{
			clearOfflineEvent();
			addOfflineEvent();
		}
		offlineDescription();
	}

	client = 0;
}

void Proxy::onHubConnected(Hub* hub_)
{
	clearEvent(hub_conn_event);

	hub_buf.clear();
	hub = hub_;

	hubConnected();
}

void Proxy::onHubLine(const std::string& line)
{
	parseCommands(hub_buf, line, &Proxy::hubCommand);
}

void Proxy::onHubDisconnected()
{
	hub = 0;

	hub_conn_event = AddEvent(60, 0);
}

void Proxy::OnEvent(int id)
{
	if (id == hub_conn_event)
	{
		handler.connectHub(profile.port);
	}

	else if (id == offline_event)
	{
		addOfflineEvent();

		save("offline_hours", toString(++offline_hours));
		offlineDescription();
	}
}

void Proxy::parseCommands(std::string& buf, const std::string& line, onCommand_type p_onCommand)
{
	char sep = separator();

	buf += line;
	if (*(buf.end() - 1) != sep)
	{
		return;
	}

	size_t i;
	while ((i = buf.find(sep)) != std::string::npos)
	{
		(this->*p_onCommand)(buf.substr(0, i));
		buf.erase(0, i + 1);
	}
}

void Proxy::clearEvent(long& id)
{
	if (id)
	{
		RemoveEvent(id);
		id = 0;
	}
}

void Proxy::offlineDescription()
{
	std::string offline_description = profile.offline_description;
	if (offline_percent != std::string::npos)
	{
		offline_description.erase(offline_percent, 1);
		offline_description.insert(offline_percent, toString(offline_hours));
	}
	offlineDescription(offline_description);
}

void Proxy::addOfflineEvent()
{
	offline_event = AddEvent(3600, 0);
}

void Proxy::clearOfflineEvent()
{
	clearEvent(offline_event);

	if (client)
	{
		offline_hours = 0;
		save("offline_hours", std::string());
	}
}
