/*
DCBouncer: bouncer / proxy for Direct Connect networks.
Copyright (C) 2008-2017 poy <poy@123gen.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "main.h"

#include <fstream>
#include <signal.h>
#include <syslog.h>

#include "Sockets/ListenSocket.h"

#include "Client.h"
#include "Crypto.h"
#include "EventHandler.h"
#include "FileSystem.h"
#include "Math.h"
#include "Profile.h"

const char* name = "DCBouncer";
bool redo_main = false;
std::auto_ptr<EventHandler> handler;

int run(const std::string& config)
{
	syslog(LOG_INFO, "start");
	FileSystem::init();
	Math::init();

	handler.reset(new EventHandler);

	{
		const std::string config_abs = FileSystem::absolute(config);
		std::ifstream f(config_abs.c_str());
		if (!f.is_open())
		{
			syslog(LOG_ERR, "can't read the settings file (path: %s)", config_abs.c_str());
			return 1;
		}

		typedef sockets::ListenSocket<Client> listener_type;
		Profile profile;

		while (!f.eof())
		{
			std::string line;
			std::getline(f, line);
			if (line.empty() || line[0] == '#')
			{
				continue;
			}
			std::string::iterator last = line.end() - 1;
			if (*last == '\r')
			{
				line.erase(last);
			}

			if (line == "OK")
			{
				bool ok = true;
				if (!profile.port)
				{
					syslog(LOG_WARNING, "invalid listening port number");
					ok = false;
				}
				if (profile.address.empty())
				{
					syslog(LOG_WARNING, "invalid hub address");
					ok = false;
				}
				if (!profile.hub_port)
				{
					syslog(LOG_WARNING, "invalid hub port number");
					ok = false;
				}

				if (ok)
				{
					listener_type* listener = new listener_type(*handler);
					if (listener->Bind(profile.port))
					{
						syslog(LOG_WARNING, "can't use port %u", profile.port);
					}
					else
					{
						handler->add(listener, profile);
						syslog(LOG_INFO, "listening on port %u for incoming connections", profile.port);
					}
				}

				profile = Profile();
				continue;
			}

			size_t i = line.find('=');
			if (i == std::string::npos)
			{
				continue;
			}

			std::string code = line.substr(0, i);
			std::string param = line.substr(i + 1);
			if (code.empty() || param.empty())
			{
				continue;
			}

			if (code == "AUTHORIZED")
			{
				sockets::Ipv4Address ip(param, 0);
				if (ip.IsValid())
				{
					Client::authorized.push_back(ip.Convert());
				}
				else
				{
					syslog(LOG_WARNING, "address %s is invalid, ignoring", param.c_str());
				}
			}

			else if (code == "CERT_FILE")
			{
				Crypto::setCertFile(param);
			}

			else if (code == "LISTENING_PORT")
			{
				profile.port = atoi(param.c_str());
			}

			else if (code == "HUB_ADDRESS")
			{
				profile.address = param;
			}

			else if (code == "HUB_PORT")
			{
				profile.hub_port = atoi(param.c_str());
			}

			else if (code == "PASS")
			{
				profile.pass = param;
			}

			else if (code == "OFFLINE_DESCRIPTION")
			{
				profile.offline_description = param;
			}
		}
	}

	handler->EventLoop();

	handler.reset();
	syslog(LOG_INFO, "stop");
	return 0;
}

int main(int argc, char* argv[])
{
	openlog(name, LOG_PID, LOG_USER);
	if (daemon(1, 0))
	{
		syslog(LOG_WARNING, "can't daemonize: %m");
	}
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, stop);
	signal(SIGTERM, stop);

	const std::string config((argc >= 2) ? argv[1] : "../etc/DCBouncer");

	do
	{
		redo_main = false;
		int res = run(config);
		if (res)
		{
			return res;
		}
	}
	while (redo_main);
	return 0;
}

void restart()
{
	redo_main = true;
	stop();
}

void stop(int)
{
	handler->SetQuit();
}
